/*
 * SISTERS.C
 *
 * Combining EGI and AniSprite to create interactive, cartoon style animations.
 * It optionally uses the Maximum MIDI toolkit to synchronize the animation
 * to the MIDI song.
 *
 * Compiling:
 *   Borland C++, 16-bit
 *      bcc -W sisters.c as16.lib eplay.lib
 *
 *   Borland C++, 32-bit
 *      bcc32 -tW sisters.c as32b.lib eplay32b.lib
 *
 *   Microsoft C/C++, 16 bit
 *      cl -Gw sisters.c as16.lib eplay.lib slibcew.lib libw.lib sisters.def
 *      rc sisters.rc sisters.exe
 *
 *   Microsoft C/C++, 32 bit
 *      cl -GA sisters.c as32m.lib eplay32m.lib user32.lib gdi32.lib
 *
 *   Watcom C/C++ 11.0, 32-bit
 *      wcl386 -l=nt_win sisters.c as32w.lib eplay32w.lib
 *
 * Copyright (c) 1997-2001, ITB CompuPhase. You may use/modify/distribute
 * this file or portions of it. It is provided as an example for the use of
 * the EGI and AniSprite APIs. There are no waranties on the correct behavior
 * of this program.
 */
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <mmsystem.h>
#include <include/anispri.h>
#include <include/eplay.h>
#include "sisters.h"

#define MAXMIDI
#if defined MAXMIDI
  #include <include/maxmidi.h>
  #include <include/mxdll.h>   /* this program uses an "internal" definition */
#endif

#if !defined QHTM_CLASSNAME
  /* these definitions were copied from QHTM.H which is part of the
   * QHTMLIGHT.DLL distribution
   */
  #if !defined _T
    #define _T(s)               s
  #endif
  #define QHTM_CLASSNAME        _T("QHTM_Window_Class_001")
  #define QHTM_LOAD_FROM_FILE   ( WM_USER + 2 )
#endif


#define BOARD_WIDTH     640
#define BOARD_HEIGHT    480

#define TEMPO           92              // approximate speed of MIDI song
#define MIDISONG        "lambada.mid"   //"bolero3.mid"


LRESULT CALLBACK AnimWinFunc(HWND hwnd, unsigned message, WPARAM wParam, LPARAM lParam);

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nCmdShow)
{
  MSG msg;
  WNDCLASS wc;
  RECT rect;
  DWORD dwStyle;
  int x, y, width, height;

  wc.style = 0;
  wc.lpfnWndProc = (WNDPROC)AnimWinFunc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon(hInstance, "sisters");
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = GetStockObject(LTGRAY_BRUSH);
  wc.lpszMenuName = (LPSTR)NULL;
  wc.lpszClassName = "Animation";
  if (!RegisterClass(&wc))
    return FALSE;

  /* create a window with the right size for the board */
  SetRect(&rect, 0, 0, BOARD_WIDTH, BOARD_HEIGHT);
  dwStyle = WS_POPUPWINDOW | WS_VISIBLE | WS_CLIPCHILDREN;
  if (GetSystemMetrics(SM_CYSCREEN)>500)
    dwStyle |= WS_CAPTION;
  AdjustWindowRect(&rect, dwStyle, FALSE);
  width = rect.right - rect.left;
  height = rect.bottom - rect.top;
  x = (GetSystemMetrics(SM_CXSCREEN) - width) / 2;
  y = (GetSystemMetrics(SM_CYSCREEN) - height) / 2;
  CreateWindow("Animation",
               "The charming sisters Fanni and Tila dance for you",
               dwStyle, x, y, width, height, 0, 0, hInstance, NULL);

  while (GetMessage(&msg, NULL, 0, 0)) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  } /* while */

  return msg.wParam;
}

static int PlayWave(LPSTR filename,int index)
{
  LPSTR lpPtr;

  if ((lpPtr=strchr(filename,'-'))!=NULL) {
    int max=atoi(lpPtr+1);
    if (max>0) {
      *(lpPtr+1)=(char)(index+'1');
      index=(index+1) % max;
    } /* if */
  } /* if */
  sndPlaySound(filename,SND_ASYNC);
  return index;
}

static int MarkerChannel = -1;

#if defined MAXMIDI
/* The Maximum MIDI toolkit comes with C++ classes, but this example is in
 * plain C. The following routines as a wrapper for three of the four function
 * blocks: MIDI file parsing, MIDI output and a MIDI timer ("sync" device).
 */
typedef struct {
  HSMF smf;
  HSYNC sync;
  HMOUT mout;
  LPMIDIEVENT events;
  int size;     /* event buffer size */
} MxMidi;

BOOL MxMidiOpen(MxMidi *mxmidi, HWND hwnd, LPCSTR filename, DWORD flags, WORD frames)
{
  HSYNC sync;
  HMOUT mout;
  HSMF smf;
  int format, tracks, qnotes;
  WORD res;
  int size;
  LPMIDIEVENT events;
  DWORD evtsize;
  LPSTR TimeSig;
  LPSTR MarkerStr;

  GetMaxMidiVersion();  /* recommended, see the Maximum MIDI book */
  memset(mxmidi, 0, sizeof(MxMidi));

  /* Allocate memory for the event buffer. Maximum MIDI does not provide
   * a macro to get the queue size from the flags, but a formula is easy
   * to come up with after examining the MAXMIDI.H file.
   */
  size = (int)(64 << ((flags & QMASK) >> 12));
  events = malloc(size * sizeof(MidiEvent));
  if (events == NULL) {
    MessageBox(hwnd, "Insufficient memory", NULL, MB_OK);
    return FALSE;
  } /* if */

  /* Open the file and check the format */
  smf = OpenSMF((LPSTR)filename, &format, 'r', &tracks);
  if (!smf) {
    free(events);
    MessageBox(hwnd, "Cannot open file (file not found)", NULL, MB_OK);
    return FALSE;
  } /* if */

  if (format!=0 || tracks!=1) {
    free(events);
    MessageBox(hwnd, "Wrong file format (not a single track format 0 file)", NULL, MB_OK);
    return FALSE;
  } /* if */

  /* Create a "sync" device and set the resolution read from the MIDI file */
  sync = OpenSync(0, hwnd, S_INT, DEFAULT_TIMERPERIOD);
  if (sync < MXMIDIERR_MAXERR) {
    free(events);
    CloseSMF(smf);
    return FALSE;
  } /* if */

  res = GetSMFResolution(smf);
  SetResolution(sync, res);

  /* Open de MIDI output; the synchronization messages and the messages to
   * keep the queue filled sent to the window whose handle is passed. If
   * no MIDI mapper is present, try open MIDI device 0.
   */
  mout = OpenMidiOut(hwnd, MIDI_MAPPER, sync, QUEUE_1K);
  if (mout < MXMIDIERR_MAXERR)
    mout = OpenMidiOut(hwnd, 0, sync, QUEUE_1K);
  if (mout < MXMIDIERR_MAXERR) {
    free(events);
    CloseSMF(smf);
    CloseSync(sync);
    return FALSE;
  } /* if */

  /* read the time signature and set the beat ratio */
  if (ReadMetaEvent(smf, 0, META_TIME_SIG, &TimeSig, &evtsize) != (DWORD)-1) {
    int num = TimeSig[0];
    int denom = 1 << TimeSig[1];
    qnotes = 4*num / denom;
  } else {
    qnotes = 4; /* no time signature, assume 4 quarter notes per measure */
  } /* if */
  SetBeatRatio(sync, (WORD)(qnotes*frames), (WORD)qnotes);

  /* see if there is a "marker channel" defined */
  while (ReadMetaEvent(smf, 0, META_TEXT, &MarkerStr, &evtsize) != (DWORD)-1) {
    if (evtsize>6 && strnicmp(MarkerStr,"@MarkerChan=",12)==0)
      MarkerChannel=atoi(MarkerStr+12) - 1;
  } /* if */

  /* Everything succeeded, copy the values before returning */
  mxmidi->smf = smf;
  mxmidi->sync = sync;
  mxmidi->mout = mout;
  mxmidi->events = events;
  mxmidi->size = size;
  return TRUE;
}

void MxMidiClose(MxMidi *mxmidi)
{
  if (mxmidi->sync) {
    StopSync(mxmidi->sync);
    CloseSync(mxmidi->sync);
  } /* if */
  if (mxmidi->mout)
    CloseMidiOut(mxmidi->mout);
  if (mxmidi->smf)
    CloseSMF(mxmidi->smf);
  if (mxmidi->events)
    free(mxmidi->events);

  memset(mxmidi, 0, sizeof(MxMidi));
}

BOOL MxMidiPlay(MxMidi *mxmidi, BOOL start)
{
  /* Call this function when you want to start playing, and then after each
   * OUTBUFFER_READY message
   */
  int num, i;

  if (!mxmidi->mout)
    return FALSE;

  num = (int)ReadSMF(mxmidi->smf, 0, mxmidi->events, mxmidi->size / 2);
  /* This function gets called when the buffer drops below 25% full.
   * In other words, the buffer is 75% empty. The line above read
   * in half the size of the buffer (so 50%). So the number of events
   * read in should always fit.
   */
  for (i=0; i < num; i++) {
    /* convert "Note On" messages in the marker channel to "user messages" */
    if (mxmidi->events[i].status==0x90+MarkerChannel)
      mxmidi->events[i].status=1;
    PutMidiOut(mxmidi->mout, &mxmidi->events[i]);
  } /* for */
  if (start)
    StartSync(mxmidi->sync);

  return TRUE;
}

#else

typedef DWORD MxMidi;
#define QUEUE_1K        0

BOOL MxMidiOpen(MxMidi *mxmidi, HWND hwnd, LPCSTR filename, DWORD flags, WORD frames)
{
  char str[256];

  wsprintf(str, "open %s type sequencer alias song wait", filename);
  *mxmidi = mciSendString(str,NULL,0,NULL);
  return *mxmidi == 0L;
}

void MxMidiClose(MxMidi *mxmidi)
{
  if (*mxmidi == 0L)
    mciSendString("close song wait",NULL,0,NULL);
  *mxmidi = 0L;
}

BOOL MxMidiPlay(MxMidi *mxmidi, BOOL start)
{
  if (*mxmidi == 0L && start)
    mciSendString("play song notify",NULL,0,NULL);
  return TRUE;
}
#endif

enum {
  MBALL,
  FANNI,
  TILA,
  MARQUEE,
  RPLAYER,
  /* --- */
  ANIMS
};
enum {
  RECORDCOVER = ANIMS,
  /* --- */
  SPRITES
};

static LPFLIC Flic[ANIMS];
static ASPRITE Sprite[SPRITES];
static int Layers[SPRITES]={1,2,3,0,0,0};
static int MaskType[ANIMS] = { AS_MASK_CUTOUT, AS_MASK_LUMA, AS_MASK_LUMA,
                               AS_MASK_LUMA, AS_MASK_CUTOUT };
static int MusicSection = 0;

static void scriptFanni(ASPRITE sprite, LPFLIC flic, int flags, int start_measure)
{
  int segment = FlicGetParam(flic,FLIC_PARAM_SEGLABEL);
  int state = (int)FlicGetData(flic,FLIC_DATA_USERDATA);
  int x,y;

  if ((flags & FLIC_FRAME_LAST)!=0) {
    switch (segment) {
    case WIGGLE1:
    case WIGGLE1_ALT:
      if (MusicSection==1) {
        state=WIGGLE2;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      } else if (state==SLIDE) {
        int nx,ny;
        /* see if she has come close enough to her rival */
        as_GetPos(sprite,&x,&y);
        as_GetPos(Sprite[TILA],&nx,&ny);
        x -= nx+80;
        if (x>10) {
          static int lookaside;
          x=10;
          if (lookaside-- < 0) {
            FlicSetParam(flic,FLIC_PARAM_SEGLABEL,WIGGLE1_ALT);
            lookaside=rand() % 3;
          } else {
            FlicSetParam(flic,FLIC_PARAM_SEGLABEL,WIGGLE1);
          } /* if */
        } else if (x<=0) {
          state=PUSH;
          FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
        } /* if */
        as_Move(sprite,-x,0);
      } else if (FlicGetParam(flic,FLIC_PARAM_CURFRAME)==1 && !start_measure) {
        /* wait for "the beat" */
        FlicSetParam(flic,FLIC_PARAM_CURFRAME,0);
      } else {
        int tstate;
        static int count;
        static BOOL bowed = FALSE;
        /* see what her rival is doing */
        tstate = (int)FlicGetData(Flic[TILA],FLIC_DATA_USERDATA);
        if (tstate==WIGGLE1 || tstate==PUSH) {
          if (++count>=5) {
            if (bowed) {
              /* Fanni has "staged" her actions already, do not repeat it */
              state=SLIDE;      /* start sliding if Tila wiggles long enough */
            } else {
              /* Fanni bows forward and hints the spectator what she
               * is going to do */
              state=BOW;
              FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
              bowed=TRUE;
            } /* if */
          } /* if */
        } else {
          count=0;
        } /* if */
      } /* if */
      break;
    case WIGGLE2:
      if (MusicSection==0) {
        state=WIGGLE1;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      } /* if */
      break;
    case BOW:
      state=SLIDE;        /* start sliding towards Tila */
      FlicSetParam(flic,FLIC_PARAM_SEGLABEL,WIGGLE1);
      break;
    case PUSH:
      /* if Fanni has just pushed her rival off-screen, she hops back
       * to her initial position.
       */
      {
        int nx,ny;
        as_GetPos(Sprite[TILA],&nx,&ny);
        state= (nx < -50) ? TURN : WIGGLE1;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      }
      break;
    case HOP:
      /* see if she has arrived on her initial position */
      as_GetPos(sprite,&x,&y);
      if (x>=330) {
        state=BACK;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      } /* if */
      break;
    default:
      /* if this segment is not "repeatable" (no ring frame), jump to
       * the next segment.
       */
      if (!FlicGetParam(flic,FLIC_PARAM_LOOP)) {
        state=FlicGetParam(flic,FLIC_PARAM_NEXTSEGMENT);
        FlicSetParam(flic,FLIC_PARAM_CURSEGMENT,state);
      } /* if */
    } /* switch */
  } /* if */

  /* labels denote that a sprite should move in some direction */
  if ((flags & FLIC_FRAME_LABEL)!=0) {
    int label=FlicGetParam(flic,FLIC_PARAM_LABEL);
    switch (label) {
    case HOP:
      as_GetPos(sprite,&x,&y);
      x = 330-x;
      if (x>10)
        x=10;
      as_Move(sprite,x,0);
      break;
    } /* switch */

  } /* if */
  if ((flags & FLIC_FRAME_USERSTRING)!=0) {
    switch (state) {
    case WIGGLE1:
      break;
    case PUSH:
      if ((int)FlicGetData(Flic[TILA],FLIC_DATA_USERDATA) == JUMP)
        PlayWave(FlicGetData(flic,FLIC_DATA_USERSTRING),0);
      break;
    default:
      PlayWave(FlicGetData(flic,FLIC_DATA_USERSTRING),0);
    } /* switch */
  } /* if */

  FlicSetData(flic,FLIC_DATA_USERDATA,(void FAR *)(DWORD)state);
}

static void scriptTila(ASPRITE sprite, LPFLIC flic, int flags, int start_measure)
{
static int JumpCount=0;
  int segment = FlicGetParam(flic,FLIC_PARAM_SEGLABEL);
  int state = (int)FlicGetData(flic,FLIC_DATA_USERDATA);
  int x,y;

  if ((flags & FLIC_FRAME_LAST)!=0) {
    switch (segment) {
    case HOP:
      if (state==HOP) {
        int zx,zy;
        /* see if she has come close enough to her rival */
        as_GetPos(sprite,&x,&y);
        as_GetPos(Sprite[FANNI],&zx,&zy);
        x = zx - x - 160;
        if (x<=0) {
          state=BACK;
          FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
        } /* if */
      } else {
        int fstate;
        static int count;
        static int wigglecount=5;
        /* see if Fanni is wiggling */
        fstate = (int)FlicGetData(Flic[FANNI],FLIC_DATA_USERDATA);
        if (fstate==WIGGLE1) {
          if (++count>=wigglecount) { /* Fanni has wiggled alone for long enough */
            state=HOP;
            wigglecount+=20;          /* wait longer before re-appearing nex time */
          } /* if */
        } else {
          count=0;
        } /* if */
      } /* if */
      break;
    case WIGGLE1:
      if (MusicSection==1) {
        state=WIGGLE2;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      } else if ((int)FlicGetData(Flic[FANNI],FLIC_DATA_USERDATA)==PUSH) {
        state= (JumpCount==2) ? PUSH : JUMP;
        JumpCount++;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      } /* if */
      break;
    case WIGGLE2:
      if (MusicSection==0) {
        state=WIGGLE1;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      } /* if */
      break;
    case BACK:
      /* if we are off-screen, go to the initial state */
      as_GetPos(sprite,&x,&y);
      if (x<-100) {
        state=NONE;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,HOP);
      } /* if */
      break;
    case PUSH:
      state=((int)FlicGetData(Flic[FANNI],FLIC_DATA_USERDATA)==PUSH) ? JUMP : WIGGLE1;
      FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      break;
    default:
      /* if this segment is not "repeatable" (no ring frame), jump to
       * the next segment.
       */
      if (!FlicGetParam(flic,FLIC_PARAM_LOOP))
        FlicSetParam(flic,FLIC_PARAM_CURSEGMENT,
                     FlicGetParam(flic,FLIC_PARAM_NEXTSEGMENT));
    } /* switch */
  } /* if */

  /* labels denote that a sprite should move in some direction */
  if ((flags & FLIC_FRAME_LABEL)!=0) {
    int label=FlicGetParam(flic,FLIC_PARAM_LABEL);
    switch (label) {
    case HOP:
      if (state==HOP) {
        int zx,zy;
        /* see if she has come close enough to her rival */
        as_GetPos(sprite,&x,&y);
        as_GetPos(Sprite[FANNI],&zx,&zy);
        x = zx - x - 160;
        if (x>10)
          x=10;
        as_Move(sprite,x,0);
      } /* if */
      break;
    case JUMP:
      /* if already nearly off-screen, move off screen faster */
      as_GetPos(sprite,&x,&y);
      as_Move(sprite,(x>=0) ? -12 : -16,0);
      break;
    } /* switch */
  } /* if */

  /* if currently stopped (because of synchronization) see whether (and
   * where) to restart.
   */
  if ((flags & FLIC_FRAME_DECODED)==0) {
    switch (segment) {
    case BACK:
      if (FlicGetParam(Flic[FANNI],FLIC_PARAM_CURFRAME)==FlicGetParam(Flic[FANNI],FLIC_PARAM_NUMFRAMES)) {
        state=WIGGLE1;
        FlicSetParam(flic,FLIC_PARAM_SEGLABEL,state);
      } /* if */
      break;
    } /* switch */
  } /* if */

  if ((flags & FLIC_FRAME_USERSTRING)!=0) {
    switch (state) {
    case BACK:
    case HOP:
    case PUSH:
      PlayWave(FlicGetData(flic,FLIC_DATA_USERSTRING),0);
      break;
    case JUMP:
      /* don't play sound when off-screen */
      as_GetPos(sprite,&x,&y);
      if (x>=-100) {
        static int waveindex=0;
        waveindex=PlayWave(FlicGetData(flic,FLIC_DATA_USERSTRING),waveindex);
      } /* if */
      break;
    } /* switch */
  } /* if */

  FlicSetData(flic,FLIC_DATA_USERDATA,(void FAR *)(DWORD)state);
}

static void scriptBall(ASPRITE sprite, LPFLIC flic, int flags, int start_measure)
{
}

static void scriptMarquee(ASPRITE sprite, LPFLIC flic, int flags, int start_measure)
{
  /* if currently stopped (because of synchronization) see whether to restart.
   */
  if ((flags & FLIC_FRAME_DECODED)==0) {
    if (FlicGetParam(Flic[FANNI],FLIC_PARAM_CURSEGMENT)==0)
      FlicSetParam(flic,FLIC_PARAM_CURSEGMENT,0);
  } /* if */

  if ((flags & FLIC_FRAME_LABEL)!=0) {
    int label=FlicGetParam(flic,FLIC_PARAM_LABEL);
    switch (label) {
    case DOWNLEVEL:
      as_Move(sprite,0,STEPSIZE);
      break;
    case UPLEVEL:
      as_Move(sprite,0,-STEPSIZE);
      break;
    } /* switch */
  } /* if */
}

static void scriptRPlayer(ASPRITE sprite, LPFLIC flic, int flags, int start_measure)
{
}

LRESULT CALLBACK AnimWinFunc(HWND hwnd, unsigned message, WPARAM wParam, LPARAM lParam)
{
static ASBOARD Board;
static MxMidi mxmidi;
  LPBITMAPINFO lpBitsInfo;
  LPVOID lpLuma;
  LPVOID lpBits;
  PAINTSTRUCT ps;
  HDC hdc;
  int width, height;
  int flags, anim, i;
  int start_measure;
  LPRECT lprc;

  switch (message) {
  case WM_CREATE:
    /* create the board (with an identity palette) */
    lpBitsInfo = as_AllocResource(sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD));
    lpBits = as_LoadDIB("checkers.bmp", lpBitsInfo, NULL, NULL);
    Board = as_CreateBoard(lpBitsInfo, lpBits, NULL,
                           AS_MODE_PAL_COLORS | AS_MODE_LUMA);
    as_SetPalette(Board, lpBitsInfo->bmiColors, TRUE);

    /* add 2 luma levels to simulate shadows; hard shadows at 30% "darkness"
     * and soft shadows at 20% "darkness" */
    lpLuma = as_CreateLumaLevel(lpBitsInfo->bmiColors, RGB(0,0,0), 77); /* 30% luma */
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 1, lpLuma);
    lpLuma = as_CreateLumaLevel(lpBitsInfo->bmiColors, RGB(0,0,0), 50); /* 20% luma */
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 2, lpLuma);
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 3, lpLuma);
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 4, lpLuma);
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 5, lpLuma);

    /* add 3 luma levels to a halo around the text (captions): 16% white,
     * 30%, 50%  */
    lpLuma = as_CreateLumaLevel(lpBitsInfo->bmiColors, RGB(255,255,255), 120);
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 3, lpLuma);
    lpLuma = as_CreateLumaLevel(lpBitsInfo->bmiColors, RGB(255,255,255), 80);
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 4, lpLuma);
    lpLuma = as_CreateLumaLevel(lpBitsInfo->bmiColors, RGB(255,255,255), 40);
    as_SetBoardData(Board, AS_DATA_LUMA_LEVEL, 5, lpLuma);
    as_FreeResource(lpBitsInfo);

    /* load the animations */
    Flic[MBALL] = FlicOpen("mball.flc");
    Flic[FANNI] = FlicOpen("fanni.flc");
    Flic[TILA] = FlicOpen("tila.flc");
    Flic[MARQUEE] = FlicOpen("marquee.flc");
    Flic[RPLAYER] = FlicOpen("rplayer.flc");

    if (Flic[FANNI]==NULL || Flic[TILA]==NULL || Flic[MARQUEE]==NULL) {
      MessageBox(hwnd,"Required animations are missing (any of the files"
                 "FANNI.FLC, TILA.FLC or MARQUEE.FLC is absent or corrupt)",
                 NULL,MB_OK);
      DestroyWindow(hwnd);
      return 0L;
    } /* if */

    /* create the sprites, using the very first frames */
    for (anim=0; anim<ANIMS; anim++) {
      if (Flic[anim]!=NULL) {
        FlicNextFrame(Flic[anim]);
        FlicSetParam(Flic[anim],FLIC_PARAM_CURFRAME,0);
        Sprite[anim] = as_Create(FlicGetParam(Flic[anim],FLIC_PARAM_WIDTH),
                                FlicGetParam(Flic[anim],FLIC_PARAM_HEIGHT),
                                FlicGetData(Flic[anim],FLIC_DATA_MASK), MaskType[anim],
                                FlicGetData(Flic[anim],FLIC_DATA_BITMAP), TRUE);
        as_Assign(Sprite[anim], Board, Layers[anim]);
      } /* if */
    } /* for */
    /* add the fixed sprites */
    Sprite[RECORDCOVER] = as_Load("reccover.bmp", RGB(255,251,240), NULL, NULL);
    as_Assign(Sprite[RECORDCOVER], Board, Layers[RECORDCOVER]);

    /* position and show the sprites */
    if (Sprite[MBALL]!=NULL)
      as_SetPos(Sprite[MBALL], 257, 167);
    if (Sprite[FANNI]!=NULL)
      as_SetPos(Sprite[FANNI], 330, 205);       // 220
    if (Sprite[TILA]!=NULL)
      as_SetPos(Sprite[TILA], -160, 205);
    if (Sprite[MARQUEE]!=NULL)
      as_SetPos(Sprite[MARQUEE], 0, -40);
    if (Sprite[RPLAYER]!=NULL)
      as_SetPos(Sprite[RPLAYER], 430, 314);
    if (Sprite[RECORDCOVER]!=NULL)
      as_SetPos(Sprite[RECORDCOVER], 472, 429);
    for (anim=0; anim<SPRITES; anim++)
      if (Sprite[anim]!=NULL)
        as_Show(Sprite[anim], TRUE);

    /* launch the MIDI file */
    #if defined MIDISONG
      MxMidiOpen(&mxmidi, hwnd, MIDISONG, QUEUE_1K, 6);
      MxMidiPlay(&mxmidi, TRUE);
    #endif

    /* Create a timer to move the sprite, set initial direction */
    #if !defined MAXMIDI
      as_SetTimer(Board, hwnd, SPEED);
    #endif
    break;

  case WM_DESTROY:
    #if defined MIDISONG
      MxMidiClose(&mxmidi);
    #endif
    /* the sprites and FLICs */
    for (anim=0; anim<ANIMS; anim++) {
      if (Sprite[anim]!=NULL) {
        as_Assign(Sprite[anim], Board, -1);
        as_Delete(Sprite[anim], FALSE);
      } /* if */
      if (Flic[anim]!=NULL)
        FlicClose(Flic[anim]);
    } /* for */
    for ( ; anim<SPRITES; anim++) {
      if (Sprite[anim]!=NULL) {
        as_Assign(Sprite[anim], Board, -1);
        as_Delete(Sprite[anim], TRUE);
      } /* if */
    } /* for */
    /* the luma levels */
    for (i = 1; i <= 5; i++) {
      lpLuma = as_GetBoardData(Board, AS_DATA_LUMA_LEVEL, i);
      as_FreeResource(lpLuma);
    } /* for */
    as_DeleteBoard(Board, TRUE);
    hdc = GetDC(hwnd);
    SetSystemPaletteUse(hdc, SYSPAL_STATIC);
    ReleaseDC(hwnd, hdc);
    PostQuitMessage(0);
    break;

#if defined MAXMIDI
  case MIDI_BEAT:
    start_measure = wParam;
#else
  case AS_TIMER:
    start_measure = 0;
#endif
    for (anim=0; anim<ANIMS; anim++) {
      if (Sprite[anim]==NULL)
        continue;
      lprc=as_GetData(Sprite[anim],AS_DATA_BOX,0);     /* get sprite position*/
      width=lprc->right - lprc->left;
      height=lprc->bottom - lprc->top;
      flags =FlicNextFrame(Flic[anim]);
      if (flags & FLIC_FRAME_IMG) {
        as_Animate(Sprite[anim],width,height,
                   FlicGetData(Flic[anim],FLIC_DATA_MASK), MaskType[anim],
                   FlicGetData(Flic[anim],FLIC_DATA_BITMAP), TRUE,
                   NULL);
      } /* if */
      switch (anim) {
      case MBALL:
        scriptBall(Sprite[anim], Flic[anim], flags, start_measure);
        break;
      case FANNI:
        scriptFanni(Sprite[anim], Flic[anim], flags, start_measure);
        break;
      case TILA:
        scriptTila(Sprite[anim], Flic[anim], flags, start_measure);
        break;
      case MARQUEE:
        scriptMarquee(Sprite[anim], Flic[anim], flags, start_measure);
        break;
      case RPLAYER:
        scriptRPlayer(Sprite[anim], Flic[anim], flags, start_measure);
        break;
      } /* switch */
    } /* for */

    /* repaint the board */
    hdc = GetDC(hwnd);
    as_PaintBoard(hdc, Board, 0, 0, FALSE);
    ReleaseDC(hwnd, hdc);
    break;

  case WM_PAINT:
    hdc=BeginPaint(hwnd, &ps);
    as_PaintBoard(hdc, Board, 0, 0, TRUE);
    EndPaint(hwnd, &ps);
    break;

  case WM_ERASEBKGND:
  case WM_PALETTECHANGED:
  case WM_QUERYNEWPALETTE:
    return as_ForwardMessage(Board, hwnd, message, wParam, lParam);

#if defined MAXMIDI
  case OUTBUFFER_READY:
    MxMidiPlay(&mxmidi, FALSE);
    break;

  case SYNC_DONE:
    MxMidiClose(&mxmidi);
    {
      HWND hwndHelp;
      WinHelp(hwnd, "sisters.hlp", HELP_CONTENTS, 0);
      hwndHelp = FindWindow("MS_WINHELP",NULL/*"Sisters"*/);
      if (IsWindow(hwndHelp)) {
        int fw = GetSystemMetrics(SM_CXFRAME);
        int fh = GetSystemMetrics(SM_CYFRAME);
        int ch = GetSystemMetrics(SM_CYCAPTION);
        SetMenu(hwndHelp, NULL);
        SetParent(hwndHelp, hwnd);
        MoveWindow(hwndHelp, -fw, -(ch+fh), BOARD_WIDTH+2*fw, BOARD_HEIGHT+ch+2*fh, TRUE);
      } /* if */
    }
    break;

  case USERMSG_BASE+1:
    MusicSection=HIWORD(lParam);
    break;
#endif

  default:
    return DefWindowProc(hwnd, message, wParam, lParam);
  } /* switch */

  return 0L;
}

